@@ -3,7 +3,7 @@
    DHCP Client. */
 
 /*
- * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
+ * Copyright (c) 2004-2008 by Internet Systems Consortium, Inc. ("ISC")
  * Copyright (c) 1995-2003 by Internet Software Consortium
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -32,7 +32,7 @@
 
 #ifndef lint
 static char ocopyright[] =
-"$Id: dhclient.c,v 1.129.2.19 2004/07/10 00:11:16 dhankins Exp $ Copyright (c) 2004 Internet Systems Consortium.  All rights reserved.\n";
+"$Id: dhclient.c,v 1.143.2.10 2008/06/11 20:20:31 dhankins Exp $ Copyright (c) 2004-2008 Internet Systems Consortium.  All rights reserved.\n";
 #endif /* not lint */
 
 #include "dhcpd.h"
@@ -61,10 +61,10 @@ struct in_addr giaddr;
    assert (state_is == state_shouldbe). */
 #define ASSERT_STATE(state_is, state_shouldbe) {}
 
-static char copyright[] = "Copyright 2004 Internet Systems Consortium.";
+static char copyright[] = "Copyright 2004-2008 Internet Systems Consortium.";
 static char arr [] = "All rights reserved.";
 static char message [] = "Internet Systems Consortium DHCP Client";
-static char url [] = "For info, please visit http://www.isc.org/products/DHCP";
+static char url [] = "For info, please visit http://www.isc.org/sw/dhcp/";
 
 u_int16_t local_port=0;
 u_int16_t remote_port=0;
@@ -77,12 +77,11 @@ int nowait=0;
 
 static void usage PROTO ((void));
 
-void do_release(struct client_state *);
-
 int main (argc, argv, envp)
 	int argc;
 	char **argv, **envp;
 {
+	int fd;
 	int i;
 	struct servent *ent;
 	struct interface_info *ip;
@@ -91,6 +90,7 @@ int main (argc, argv, envp)
 	char *server = (char *)0;
 	char *relay = (char *)0;
 	isc_result_t status;
+ 	int exit_mode = 0;
  	int release_mode = 0;
 	omapi_object_t *listener;
 	isc_result_t result;
@@ -102,15 +102,18 @@ int main (argc, argv, envp)
 	int no_dhclient_script = 0;
 	char *s;
 
-	/* Make sure we have stdin, stdout and stderr. */
-	i = open ("/dev/null", O_RDWR);
-	if (i == 0)
-		i = open ("/dev/null", O_RDWR);
-	if (i == 1) {
-		i = open ("/dev/null", O_RDWR);
-		log_perror = 0; /* No sense logging to /dev/null. */
-	} else if (i != -1)
-		close (i);
+        /* Make sure that file descriptors 0 (stdin), 1, (stdout), and
+           2 (stderr) are open. To do this, we assume that when we
+           open a file the lowest available file decriptor is used. */
+        fd = open("/dev/null", O_RDWR);
+        if (fd == 0)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 1)
+                fd = open("/dev/null", O_RDWR);
+        if (fd == 2)
+                log_perror = 0; /* No sense logging to /dev/null. */
+        else if (fd != -1)
+                close(fd);
 
 #ifdef SYSLOG_4_2
 	openlog ("dhclient", LOG_NDELAY);
@@ -141,6 +144,10 @@ int main (argc, argv, envp)
 		if (!strcmp (argv [i], "-r")) {
 			release_mode = 1;
 			no_daemon = 1;
+                } else if (!strcmp (argv [i], "-x")) { /* eXit, no release */
+                        release_mode = 0;
+                        no_daemon = 0;
+                        exit_mode = 1;
 		} else if (!strcmp (argv [i], "-p")) {
 			if (++i == argc)
 				usage ();
@@ -212,10 +219,10 @@ int main (argc, argv, envp)
  		    if (status != ISC_R_SUCCESS)
  			log_fatal ("Can't record interface %s:%s",
 				   argv [i], isc_result_totext (status));
-		    if (strlen (argv [i]) > sizeof tmp -> name)
-			    log_fatal ("%s: interface name too long (max %ld)",
-				       argv [i], (long)strlen (argv [i]));
- 		    strcpy (tmp -> name, argv [i]);
+		    if (strlen(argv[i]) >= sizeof(tmp->name))
+			    log_fatal("%s: interface name too long (is %ld)",
+				       argv [i], (long)strlen(argv[i]));
+		    strcpy(tmp->name, argv[i]);
 		    if (interfaces) {
 			    interface_reference (&tmp -> next,
 						 interfaces, MDL);
@@ -241,7 +248,7 @@ int main (argc, argv, envp)
 	}
 
 	/* first kill of any currently running client */
-	if (release_mode) {
+	if (release_mode || exit_mode) {
 		FILE *pidfd;
 		pid_t oldpid;
 		long temp;
@@ -253,10 +260,8 @@ int main (argc, argv, envp)
 			oldpid = (pid_t)temp;
 
 			if (e != 0 && e != EOF) {
-				if (oldpid) {
-					if (kill(oldpid, SIGTERM) == 0)
-						unlink(path_dhclient_pid);
-				}
+				if (oldpid)
+					kill(oldpid, SIGTERM);
 			}
 			fclose(pidfd);
 		}
@@ -365,7 +370,7 @@ int main (argc, argv, envp)
 			log_info ("No broadcast interfaces found - exiting.");
 			exit (0);
 		}
-	} else if (!release_mode) {
+	} else if (!release_mode && !exit_mode) {
 		/* Call the script with the list of interfaces. */
 		for (ip = interfaces; ip; ip = ip -> next) {
 			/* If interfaces were specified, don't configure
@@ -411,7 +416,9 @@ int main (argc, argv, envp)
 	for (ip = interfaces; ip; ip = ip -> next) {
 		ip -> flags |= INTERFACE_RUNNING;
 		for (client = ip -> client; client; client = client -> next) {
-			if (release_mode)
+                        if (exit_mode)
+                                state_stop (client);
+                        else if (release_mode)
 				do_release (client);
 			else {
 				client -> state = S_INIT;
@@ -423,7 +430,7 @@ int main (argc, argv, envp)
 		}
 	}
 
-	if (release_mode)
+	if (release_mode || exit_mode)
 		return 0;
 
 	/* Start up a listener for the object management API protocol. */
@@ -476,7 +483,7 @@ static void usage ()
 	log_info (arr);
 	log_info (url);
 
-	log_error ("Usage: dhclient [-1dqr] [-nw] [-p <port>] %s",
+	log_error ("Usage: dhclient [-1dqrx] [-nw] [-p <port>] %s",
 		   "[-s server]");
 	log_error ("                [-cf config-file] [-lf lease-file]%s",
 		   "[-pf pid-file] [-e VAR=val]");
@@ -789,11 +796,9 @@ void dhcpack (packet)
 		client -> new -> renewal = TIME_MAX;
 
 	/* Now introduce some randomness to the renewal time: */
-	if (client -> new -> renewal <= TIME_MAX / 3 - 3)
-		client -> new -> renewal =
-				(((client -> new -> renewal + 3) * 3 / 4) +
-				    (random () % /* XXX NUMS */
-				     ((client -> new -> renewal + 3) / 4)));
+	if (client->new->renewal <= ((TIME_MAX / 3) - 3))
+		client->new->renewal = (((client->new->renewal * 3) + 3) / 4) +
+				(((random() % client->new->renewal) + 3) / 4);
 
 	/* Same deal with the rebind time. */
 	oc = lookup_option (&dhcp_universe, client -> new -> options,
@@ -936,6 +941,8 @@ void state_bound (cpp)
 			client -> destination.len = 4;
 		} else
 			client -> destination = iaddr_broadcast;
+
+		data_string_forget (&ds, MDL);
 	} else
 		client -> destination = iaddr_broadcast;
 
@@ -998,7 +1005,9 @@ void db_startup (testp)
 void bootp (packet)
 	struct packet *packet;
 {
-	struct iaddrlist *ap;
+	struct iaddrmatchlist *ap;
+	char addrbuf[4*16];
+	char maskbuf[4*16];
 
 	if (packet -> raw -> op != BOOTREPLY)
 		return;
@@ -1007,9 +1016,17 @@ void bootp (packet)
 	   on it. */
 	for (ap = packet -> interface -> client -> config -> reject_list;
 	     ap; ap = ap -> next) {
-		if (addr_eq (packet -> client_addr, ap -> addr)) {
-			log_info ("BOOTREPLY from %s rejected.",
-			      piaddr (ap -> addr));
+		if (addr_match(&packet->client_addr, &ap->match)) {
+
+		        /* piaddr() returns its result in a static
+			   buffer sized 4*16 (see common/inet.c). */
+
+		        strcpy(addrbuf, piaddr(ap->match.addr));
+		        strcpy(maskbuf, piaddr(ap->match.mask));
+
+			log_info("BOOTREPLY from %s rejected by rule %s "
+				 "mask %s.", piaddr(packet->client_addr),
+				 addrbuf, maskbuf);
 			return;
 		}
 	}
@@ -1021,9 +1038,11 @@ void bootp (packet)
 void dhcp (packet)
 	struct packet *packet;
 {
-	struct iaddrlist *ap;
+	struct iaddrmatchlist *ap;
 	void (*handler) PROTO ((struct packet *));
 	const char *type;
+	char addrbuf[4*16];
+	char maskbuf[4*16];
 
 	switch (packet -> packet_type) {
 	      case DHCPOFFER:
@@ -1049,9 +1068,17 @@ void dhcp (packet)
 	   on it. */
 	for (ap = packet -> interface -> client -> config -> reject_list;
 	     ap; ap = ap -> next) {
-		if (addr_eq (packet -> client_addr, ap -> addr)) {
-			log_info ("%s from %s rejected.",
-			      type, piaddr (ap -> addr));
+		if (addr_match(&packet->client_addr, &ap->match)) {
+
+		        /* piaddr() returns its result in a static
+			   buffer sized 4*16 (see common/inet.c). */
+
+		        strcpy(addrbuf, piaddr(ap->match.addr));
+		        strcpy(maskbuf, piaddr(ap->match.mask));
+
+			log_info("%s from %s rejected by rule %s mask %s.",
+				 type, piaddr(packet->client_addr),
+				 addrbuf, maskbuf);
 			return;
 		}
 	}
@@ -1104,13 +1131,22 @@ void dhcpoffer (packet)
 		if (!lookup_option
 		    (&dhcp_universe, packet -> options,
 		     client -> config -> required_options [i])) {
-		    log_info ("%s: no %s option.",
-			      obuf, (dhcp_universe.options
-				     [client -> config -> required_options [i]]
-				     -> name));
-				return;
-			}
+		    struct option *option = NULL;
+		    unsigned code = client->config->required_options[i];
+
+		    option_code_hash_lookup(&option, dhcp_universe.code_hash,
+					    &code, 0, MDL);
+
+		    if (option)
+			log_info("%s: no %s option.", obuf, option->name);
+		    else
+			log_info("%s: no unknown-%u option.", obuf, code);
+
+		    option_dereference(&option, MDL);
+
+		    return;
 		}
+	    }
 	}
 
 	/* If we've already seen this lease, don't record it again. */
@@ -1184,6 +1220,7 @@ struct client_lease *packet_to_lease (packet, client)
 	struct client_lease *lease;
 	unsigned i;
 	struct option_cache *oc;
+	struct option *option = NULL;
 	struct data_string data;
 
 	lease = (struct client_lease *)new_client_lease (MDL);
@@ -1216,11 +1253,18 @@ struct client_lease *packet_to_lease (packet, client)
 					   packet -> options, lease -> options,
 					   &global_scope, oc, MDL)) {
 			if (data.len) {
+				if (!option_code_hash_lookup(&option,
+						dhcp_universe.code_hash,
+						&i, 0, MDL))
+					log_fatal("Unable to find VENDOR "
+						  "option (%s:%d).", MDL);
 				parse_encapsulated_suboptions
-					(packet -> options, &dhcp_options [i],
+					(packet -> options, option,
 					 data.data, data.len, &dhcp_universe,
 					 client -> config -> vendor_space_name
 						);
+
+				option_dereference(&option, MDL);
 			}
 			data_string_forget (&data, MDL);
 		}
@@ -1246,12 +1290,12 @@ struct client_lease *packet_to_lease (packet, client)
 	if (!(i & 2) && packet -> raw -> sname [0]) {
 		unsigned len;
 		/* Don't count on the NUL terminator. */
-		for (len = 0; len < 64; len++)
+		for (len = 0; len < DHCP_SNAME_LEN; len++)
 			if (!packet -> raw -> sname [len])
 				break;
 		lease -> server_name = dmalloc (len + 1, MDL);
 		if (!lease -> server_name) {
-			log_error ("dhcpoffer: no memory for filename.\n");
+			log_error ("dhcpoffer: no memory for server name.\n");
 			destroy_client_lease (lease);
 			return (struct client_lease *)0;
 		} else {
@@ -1265,7 +1309,7 @@ struct client_lease *packet_to_lease (packet, client)
 	if (!(i & 1) && packet -> raw -> file [0]) {
 		unsigned len;
 		/* Don't count on the NUL terminator. */
-		for (len = 0; len < 64; len++)
+		for (len = 0; len < DHCP_FILE_LEN; len++)
 			if (!packet -> raw -> file [len])
 				break;
 		lease -> filename = dmalloc (len + 1, MDL);
@@ -1334,12 +1378,33 @@ void dhcpnak (packet)
 		return;
 	}
 
+	/* If we get a DHCPNAK, we use the EXPIRE dhclient-script state
+	 * to indicate that we want all old bindings to be removed.  (It
+	 * is possible that we may get a NAK while in the RENEW state,
+	 * so we might have bindings active at that time)
+	 */
+	script_init(client, "EXPIRE", NULL);
+	script_write_params(client, "old_", client->active);
+	if (client->alias)
+		script_write_params(client, "alias_", client->alias);
+	script_go(client);
+
 	destroy_client_lease (client -> active);
 	client -> active = (struct client_lease *)0;
 
 	/* Stop sending DHCPREQUEST packets... */
 	cancel_timeout (send_request, client);
 
+	/* On some scripts, 'EXPIRE' causes the interface to be ifconfig'd
+	 * down (this expunges any routes and arp cache).  This makes the
+	 * interface unusable by state_init(), which we call next.  So, we
+	 * need to 'PREINIT' the interface to bring it back up.
+	 */
+	script_init(client, "PREINIT", NULL);
+	if (client->alias)
+		script_write_params(client, "alias_", client->alias);
+	script_go(client);
+
 	client -> state = S_INIT;
 	state_init (client);
 }
@@ -1398,26 +1463,21 @@ void send_discover (cpp)
 
 	/* If we're supposed to increase the interval, do so.  If it's
 	   currently zero (i.e., we haven't sent any packets yet), set
-	   it to one; otherwise, add to it a random number between
-	   zero and two times itself.  On average, this means that it
-	   will double with every transmission. */
+	   it to initial_interval; otherwise, add to it a random number
+	   between zero and two times itself.  On average, this means
+	   that it will double with every transmission. */
 	if (increase) {
-		if (!client -> interval)
-			client -> interval =
-				client -> config -> initial_interval;
+		if (!client->interval)
+			client->interval = client->config->initial_interval;
 		else
-			client -> interval += ((random () >> 2) %
-					       (2 * client -> interval));
+			client->interval += random() % (2 * client->interval);
 
 		/* Don't backoff past cutoff. */
-		if (client -> interval >
-		    client -> config -> backoff_cutoff)
-			client -> interval =
-				((client -> config -> backoff_cutoff / 2)
-				 + ((random () >> 2) %
-				    client -> config -> backoff_cutoff));
-	} else if (!client -> interval)
-		client -> interval = client -> config -> initial_interval;
+		if (client->interval > client->config->backoff_cutoff)
+			client->interval = (client->config->backoff_cutoff / 2)
+				 + (random() % client->config->backoff_cutoff);
+	} else if (!client->interval)
+		client->interval = client->config->initial_interval;
 		
 	/* If the backoff would take us to the panic timeout, just use that
 	   as the interval. */
@@ -1644,7 +1704,8 @@ void send_request (cpp)
 	    client -> config -> backoff_cutoff)
 		client -> interval =
 			((client -> config -> backoff_cutoff / 2)
-			 + ((random () >> 2) % client -> interval));
+			 + ((random () >> 2) %
+					client -> config -> backoff_cutoff));
 
 	/* If the backoff would take us to the expiry time, just set the
 	   timeout to the expiry time. */
@@ -1793,6 +1854,7 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 {
 	unsigned i;
 	struct option_cache *oc;
+	struct option *option = NULL;
 	struct buffer *bp = (struct buffer *)0;
 
 	/* If there are any leftover options, get rid of them. */
@@ -1811,26 +1873,31 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 	/* Send the requested address if provided. */
 	if (rip) {
 		client -> requested_address = *rip;
-		if (!(make_const_option_cache
-		      (&oc, (struct buffer **)0, rip -> iabuf, rip -> len,
-		       &dhcp_options [DHO_DHCP_REQUESTED_ADDRESS], MDL)))
+		i = DHO_DHCP_REQUESTED_ADDRESS;
+		if (!(option_code_hash_lookup(&option, dhcp_universe.code_hash,
+					      &i, 0, MDL) &&
+		      make_const_option_cache(&oc, NULL, rip->iabuf, rip->len,
+					      option, MDL)))
 			log_error ("can't make requested address cache.");
 		else {
 			save_option (&dhcp_universe, *op, oc);
 			option_cache_dereference (&oc, MDL);
 		}
+		option_dereference(&option, MDL);
 	} else {
 		client -> requested_address.len = 0;
 	}
 
-	if (!(make_const_option_cache
-	      (&oc, (struct buffer **)0,
-	       type, 1, &dhcp_options [DHO_DHCP_MESSAGE_TYPE], MDL)))
+	i = DHO_DHCP_MESSAGE_TYPE;
+	if (!(option_code_hash_lookup(&option, dhcp_universe.code_hash, &i, 0,
+				      MDL) &&
+	      make_const_option_cache(&oc, NULL, type, 1, option, MDL)))
 		log_error ("can't make message type.");
 	else {
 		save_option (&dhcp_universe, *op, oc);
 		option_cache_dereference (&oc, MDL);
 	}
+	option_dereference(&option, MDL);
 
 	if (prl) {
 		/* Figure out how many parameters were requested. */
@@ -1839,17 +1906,21 @@ void make_client_options (client, lease, type, sid, rip, prl, op)
 		if (!buffer_allocate (&bp, i, MDL))
 			log_error ("can't make parameter list buffer.");
 		else {
+			unsigned code = DHO_DHCP_PARAMETER_REQUEST_LIST;
+
 			for (i = 0; prl [i]; i++)
 				bp -> data [i] = prl [i];
-			if (!(make_const_option_cache
-			      (&oc, &bp, (u_int8_t *)0, i,
-			       &dhcp_options [DHO_DHCP_PARAMETER_REQUEST_LIST],
-			       MDL)))
+			if (!(option_code_hash_lookup(&option,
+						      dhcp_universe.code_hash,
+						      &code, 0, MDL) &&
+			      make_const_option_cache(&oc, &bp, NULL, i,
+						      option, MDL)))
 				log_error ("can't make option cache");
 			else {
 				save_option (&dhcp_universe, *op, oc);
 				option_cache_dereference (&oc, MDL);
 			}
+			option_dereference(&option, MDL);
 		}
 	}
 
@@ -1946,6 +2017,9 @@ void make_request (client, lease)
 	else
 		oc = (struct option_cache *)0;
 
+	if (client -> sent_options)
+		option_state_dereference (&client -> sent_options, MDL);
+
 	make_client_options (client, lease, &request, oc,
 			     ((client -> state == S_REQUESTING ||
 			       client -> state == S_REBOOTING)
@@ -1968,7 +2042,6 @@ void make_request (client, lease)
 			      (struct data_string *)0,
 			      client -> config -> vendor_space_name);
 
-	option_state_dereference (&client -> sent_options, MDL);
 	if (client -> packet_length < BOOTP_MIN_LEN)
 		client -> packet_length = BOOTP_MIN_LEN;
 
@@ -2026,12 +2099,13 @@ void make_decline (client, lease)
 
 	struct option_state *options = (struct option_state *)0;
 
+	/* Create the options cache. */
 	oc = lookup_option (&dhcp_universe, lease -> options,
 			    DHO_DHCP_SERVER_IDENTIFIER);
 	make_client_options (client, lease, &decline, oc,
 			     &lease -> address, (u_int32_t *)0, &options);
 
-	/* Set up the option buffer... */
+	/* Consume the options cache into the option buffer. */
 	memset (&client -> packet, 0, sizeof (client -> packet));
 	client -> packet_length =
 		cons_options ((struct packet *)0, &client -> packet,
@@ -2039,10 +2113,12 @@ void make_decline (client, lease)
 			      (struct option_state *)0, options,
 			      &global_scope, 0, 0, 0, (struct data_string *)0,
 			      client -> config -> vendor_space_name);
+
+	/* Destroy the options cache. */
 	option_state_dereference (&options, MDL);
+
 	if (client -> packet_length < BOOTP_MIN_LEN)
 		client -> packet_length = BOOTP_MIN_LEN;
-	option_state_dereference (&options, MDL);
 
 	client -> packet.op = BOOTREQUEST;
 	client -> packet.htype = client -> interface -> hw_address.hbuf [0];
@@ -2235,6 +2311,7 @@ int write_client_lease (client, lease, rewrite, makesure)
 	pair *hash;
 	int errors = 0;
 	char *s;
+	const char *tval;
 
 	if (!rewrite) {
 		if (leases_written++ > 20) {
@@ -2297,15 +2374,15 @@ int write_client_lease (client, lease, rewrite, makesure)
 			errors++;
 
 	}
-	if (lease -> server_name) {
-		s = quotify_string (lease -> filename, MDL);
-		if (s) {
-			fprintf (leaseFile, "  server-name \"%s\";\n", s);
+	if (lease->server_name != NULL) {
+		s = quotify_string(lease->server_name, MDL);
+		if (s != NULL) {
+			fprintf(leaseFile, "  server-name \"%s\";\n", s);
 			if (errno) {
 				++errors;
 				errno = 0;
 			}
-			dfree (s, MDL);
+			dfree(s, MDL);
 		} else
 			++errors;
 	}
@@ -2336,49 +2413,27 @@ int write_client_lease (client, lease, rewrite, makesure)
 				      client, write_lease_option);
 	}
 
-	/* Note: the following is not a Y2K bug - it's a Y1.9K bug.   Until
-	   somebody invents a time machine, I think we can safely disregard
-	   it. */
-	t = gmtime (&lease -> renewal);
-	fprintf (leaseFile,
-		 "  renew %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+	tval = print_time(lease->renewal);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  renew %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	t = gmtime (&lease -> rebind);
-	fprintf (leaseFile,
-		 "  rebind %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+
+	tval = print_time(lease->rebind);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  rebind %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	t = gmtime (&lease -> expiry);
-	fprintf (leaseFile,
-		 "  expire %d %d/%d/%d %02d:%02d:%02d;\n",
-		 t -> tm_wday, t -> tm_year + 1900,
-		 t -> tm_mon + 1, t -> tm_mday,
-		 t -> tm_hour, t -> tm_min, t -> tm_sec);
-	if (errno != 0) {
+
+	tval = print_time(lease->expiry);
+	if (tval == NULL ||
+	    fprintf(leaseFile, "  expire %s\n", tval) < 0)
 		errors++;
-		errno = 0;
-	}
-	fprintf (leaseFile, "}\n");
-	if (errno != 0) {
+
+	if (fprintf(leaseFile, "}\n") < 0)
 		errors++;
-		errno = 0;
-	}
-	fflush (leaseFile);
-	if (errno != 0) {
+
+	if (fflush(leaseFile) != 0)
 		errors++;
-		errno = 0;
-	}
+
 	if (!errors && makesure) {
 		if (fsync (fileno (leaseFile)) < 0) {
 			log_info ("write_client_lease: %m");
@@ -2602,6 +2657,11 @@ int script_go (client)
 			wstatus = 0;
 		}
 	} else {
+		/* We don't want to pass an open file descriptor for
+		 * dhclient.leases when executing dhclient-script.
+		 */
+		if (leaseFile != NULL)
+			fclose(leaseFile);
 		execve (scriptName, argv, envp);
 		log_error ("execve (%s, ...): %m", scriptName);
 		exit (0);
@@ -2698,7 +2758,6 @@ void go_daemon ()
 {
 	static int state = 0;
 	int pid;
-	int i;
 
 	/* Don't become a daemon if the user requested otherwise. */
 	if (no_daemon) {
@@ -2728,14 +2787,10 @@ void go_daemon ()
         close(2);
 
 	/* Reopen them on /dev/null. */
-	i = open ("/dev/null", O_RDWR);
-	if (i == 0)
-		i = open ("/dev/null", O_RDWR);
-	if (i == 1) {
-		i = open ("/dev/null", O_RDWR);
-		log_perror = 0; /* No sense logging to /dev/null. */
-	} else if (i != -1)
-		close (i);
+	open("/dev/null", O_RDWR);
+	open("/dev/null", O_RDWR);
+	open("/dev/null", O_RDWR);
+	log_perror = 0; /* No sense logging to /dev/null. */
 
 	write_client_pid_file ();
 }
@@ -2825,6 +2880,8 @@ void do_release(client)
 				client -> destination.len = 4;
 			} else
 				client -> destination = iaddr_broadcast;
+
+			data_string_forget (&ds, MDL);
 		} else
 			client -> destination = iaddr_broadcast;
 		client -> first_sending = cur_time;
@@ -3045,7 +3102,7 @@ void client_dns_update_timeout (void *cp)
 		}
 	}
 }
-			
+
 /* See if we should do a DNS update, and if so, do it. */
 
 isc_result_t client_dns_update (struct client_state *client, int addp, int ttl)
@@ -3141,12 +3198,13 @@ isc_result_t client_dns_update (struct client_state *client, int addp, int ttl)
 			rcode = ddns_update_a (&ddns_fwd_name,
 					       client -> active -> address,
 					       &ddns_dhcid, ttl,
-					       1);
+					       1, 1);
 		else
 			rcode = ddns_remove_a (&ddns_fwd_name,
 					       client -> active -> address,
 					       &ddns_dhcid);
-	}
+	} else
+		rcode = ISC_R_FAILURE;
 	
 	data_string_forget (&ddns_fwd_name, MDL);
 	data_string_forget (&ddns_dhcid, MDL);
